@@ -11,11 +11,11 @@ const char argv0[] = "dumpelf";
 #include "paxinc.h"
 
 /* prototypes */
-static void dumpelf(const char *filename, long file_cnt);
+static void dumpelf(const char *filename, size_t file_cnt);
 static void dump_ehdr(elfobj *elf, const void *ehdr);
-static void dump_phdr(elfobj *elf, const void *phdr, long phdr_cnt);
-static void dump_shdr(elfobj *elf, const void *shdr, long shdr_cnt, const char *name);
-static void dump_dyn(elfobj *elf, const void *dyn, long dyn_cnt);
+static void dump_phdr(elfobj *elf, const void *phdr, size_t phdr_cnt);
+static void dump_shdr(elfobj *elf, const void *shdr, size_t shdr_cnt, const char *section_name);
+static void dump_dyn(elfobj *elf, const void *dyn, size_t dyn_cnt);
 #if 0
 static void dump_sym(elfobj *elf, const void *sym);
 static void dump_rel(elfobj *elf, const void *rel);
@@ -31,10 +31,10 @@ static char be_verbose = 0;
 static const void *phdr_dynamic_void;
 
 /* dump all internal elf info */
-static void dumpelf(const char *filename, long file_cnt)
+static void dumpelf(const char *filename, size_t file_cnt)
 {
 	elfobj *elf;
-	unsigned long i, b;
+	size_t i, b;
 
 	/* verify this is real ELF */
 	if ((elf = readelf(filename)) == NULL)
@@ -48,9 +48,9 @@ static void dumpelf(const char *filename, long file_cnt)
 		"\n"
 		"/*\n"
 		" * ELF dump of '%s'\n"
-		" *     %li (0x%lX) bytes\n"
+		" *     %ji (0x%jX) bytes\n"
 		" */\n\n",
-		filename, (unsigned long)elf->len, (unsigned long)elf->len);
+		filename, elf->len, elf->len);
 
 	/* setup the struct to namespace this elf */
 #define MAKE_STRUCT(B) \
@@ -58,16 +58,16 @@ static void dumpelf(const char *filename, long file_cnt)
 	Elf ## B ## _Ehdr *ehdr = EHDR ## B (elf->ehdr); \
 	b = B; \
 	printf( \
-		"Elf%1$i_Dyn dumpedelf_dyn_%2$li[];\n" \
+		"Elf%1$i_Dyn dumpedelf_dyn_%2$zu[];\n" \
 		"struct {\n" \
 		"\tElf%1$i_Ehdr ehdr;\n" \
-		"\tElf%1$i_Phdr phdrs[%3$li];\n" \
-		"\tElf%1$i_Shdr shdrs[%4$li];\n" \
+		"\tElf%1$i_Phdr phdrs[%3$u];\n" \
+		"\tElf%1$i_Shdr shdrs[%4$u];\n" \
 		"\tElf%1$i_Dyn *dyns;\n" \
-		"} dumpedelf_%2$li = {\n\n", \
+		"} dumpedelf_%2$zu = {\n\n", \
 		B, file_cnt, \
-		(long)EGET(ehdr->e_phnum), \
-		(long)EGET(ehdr->e_shnum) \
+		(uint16_t)EGET(ehdr->e_phnum), \
+		(uint16_t)EGET(ehdr->e_shnum) \
 	); \
 	}
 	MAKE_STRUCT(32)
@@ -123,11 +123,11 @@ static void dumpelf(const char *filename, long file_cnt)
 	printf("},\n");
 
 	/* finish the namespace struct and start the abitrary ones */
-	printf("\n.dyns = dumpedelf_dyn_%li,\n", file_cnt);
+	printf("\n.dyns = dumpedelf_dyn_%zu,\n", file_cnt);
 	printf("};\n");
 
 	/* start the arbitrary structs */
-	printf("Elf%lu_Dyn dumpedelf_dyn_%li[] = {\n", b, file_cnt);
+	printf("Elf%zu_Dyn dumpedelf_dyn_%zu[] = {\n", b, file_cnt);
 	if (phdr_dynamic_void) {
 #define DUMP_DYNS(B) \
 		if (elf->elf_class == ELFCLASS ## B) { \
@@ -161,48 +161,98 @@ static void dump_ehdr(elfobj *elf, const void *ehdr_void)
 	printf(".ehdr = {\n"); \
 	printf("\t.e_ident = { /* (EI_NIDENT bytes) */\n" \
 	       "\t\t/* [%i] EI_MAG:        */ 0x%X,'%c','%c','%c',\n" \
-	       "\t\t/* [%i] EI_CLASS:      */ %i , /* (%s) */\n" \
-	       "\t\t/* [%i] EI_DATA:       */ %i , /* (%s) */\n" \
-	       "\t\t/* [%i] EI_VERSION:    */ %i , /* (%s) */\n" \
-	       "\t\t/* [%i] EI_OSABI:      */ %i , /* (%s) */\n" \
-	       "\t\t/* [%i] EI_ABIVERSION: */ %i ,\n" \
+	       "\t\t/* [%i] EI_CLASS:      */ %u , /* (%s) */\n" \
+	       "\t\t/* [%i] EI_DATA:       */ %u , /* (%s) */\n" \
+	       "\t\t/* [%i] EI_VERSION:    */ %u , /* (%s) */\n" \
+	       "\t\t/* [%i] EI_OSABI:      */ %u , /* (%s) */\n" \
+	       "\t\t/* [%i] EI_ABIVERSION: */ %u ,\n" \
 	       "\t\t/* [%i-%i] EI_PAD:     */ 0x%X, 0x%X, 0x%X, 0x%X, 0x%X, 0x%X, 0x%X,\n" \
 	       "\t},\n", \
-	       EI_MAG0, (unsigned int)ehdr->e_ident[EI_MAG0], ehdr->e_ident[EI_MAG1], ehdr->e_ident[EI_MAG2], ehdr->e_ident[EI_MAG3], \
-	       EI_CLASS, (int)ehdr->e_ident[EI_CLASS], get_elfeitype(EI_CLASS, ehdr->e_ident[EI_CLASS]), \
-	       EI_DATA, (int)ehdr->e_ident[EI_DATA], get_elfeitype(EI_DATA, ehdr->e_ident[EI_DATA]), \
-	       EI_VERSION, (int)ehdr->e_ident[EI_VERSION], get_elfeitype(EI_VERSION, ehdr->e_ident[EI_VERSION]), \
-	       EI_OSABI, (int)ehdr->e_ident[EI_OSABI], get_elfeitype(EI_OSABI, ehdr->e_ident[EI_OSABI]), \
-	       EI_ABIVERSION, (int)ehdr->e_ident[EI_ABIVERSION], \
+	       EI_MAG0, ehdr->e_ident[EI_MAG0], ehdr->e_ident[EI_MAG1], ehdr->e_ident[EI_MAG2], ehdr->e_ident[EI_MAG3], \
+	       EI_CLASS, ehdr->e_ident[EI_CLASS], get_elfeitype(EI_CLASS, ehdr->e_ident[EI_CLASS]), \
+	       EI_DATA, ehdr->e_ident[EI_DATA], get_elfeitype(EI_DATA, ehdr->e_ident[EI_DATA]), \
+	       EI_VERSION, ehdr->e_ident[EI_VERSION], get_elfeitype(EI_VERSION, ehdr->e_ident[EI_VERSION]), \
+	       EI_OSABI, ehdr->e_ident[EI_OSABI], get_elfeitype(EI_OSABI, ehdr->e_ident[EI_OSABI]), \
+	       EI_ABIVERSION, ehdr->e_ident[EI_ABIVERSION], \
 	       EI_PAD, EI_NIDENT - 1, \
-	         (unsigned int)ehdr->e_ident[EI_PAD + 0], \
-	         (unsigned int)ehdr->e_ident[EI_PAD + 1], \
-	         (unsigned int)ehdr->e_ident[EI_PAD + 2], \
-	         (unsigned int)ehdr->e_ident[EI_PAD + 3], \
-	         (unsigned int)ehdr->e_ident[EI_PAD + 4], \
-	         (unsigned int)ehdr->e_ident[EI_PAD + 5], \
-	         (unsigned int)ehdr->e_ident[EI_PAD + 6] \
+	         ehdr->e_ident[EI_PAD + 0], \
+	         ehdr->e_ident[EI_PAD + 1], \
+	         ehdr->e_ident[EI_PAD + 2], \
+	         ehdr->e_ident[EI_PAD + 3], \
+	         ehdr->e_ident[EI_PAD + 4], \
+	         ehdr->e_ident[EI_PAD + 5], \
+	         ehdr->e_ident[EI_PAD + 6] \
 	); \
-	printf("\t.e_type      = %-10i , /* (%s) */\n", (int)EGET(ehdr->e_type), get_elfetype(elf)); \
-	printf("\t.e_machine   = %-10i , /* (%s) */\n", (int)EGET(ehdr->e_machine), get_elfemtype(elf)); \
-	printf("\t.e_version   = %-10i , /* (%s) */\n", (int)EGET(ehdr->e_version), get_elfeitype(EI_VERSION, EGET(ehdr->e_version))); \
-	printf("\t.e_entry     = 0x%-8lX , /* (start address at runtime) */\n", (unsigned long)EGET(ehdr->e_entry)); \
-	printf("\t.e_phoff     = %-10li , /* (bytes into file) */\n", (unsigned long)EGET(ehdr->e_phoff)); \
-	printf("\t.e_shoff     = %-10li , /* (bytes into file) */\n", (unsigned long)EGET(ehdr->e_shoff)); \
-	printf("\t.e_flags     = 0x%-8X ,\n", (unsigned int)EGET(ehdr->e_flags)); \
-	printf("\t.e_ehsize    = %-10i , /* (bytes) */\n", (int)EGET(ehdr->e_ehsize)); \
-	printf("\t.e_phentsize = %-10i , /* (bytes) */\n", (int)EGET(ehdr->e_phentsize)); \
+	printf("\t.e_type      = %-10u , /* (%s) */\n", (uint16_t)EGET(ehdr->e_type), get_elfetype(elf)); \
+	printf("\t.e_machine   = %-10u , /* (%s) */\n", (uint16_t)EGET(ehdr->e_machine), get_elfemtype(elf)); \
+	printf("\t.e_version   = %-10u , /* (%s) */\n", (uint32_t)EGET(ehdr->e_version), get_elfeitype(EI_VERSION, EGET(ehdr->e_version))); \
+	printf("\t.e_entry     = 0x%-8"PRIX64" , /* (start address at runtime) */\n", EGET(ehdr->e_entry)); \
+	printf("\t.e_phoff     = %-10"PRIi64" , /* (bytes into file) */\n", EGET(ehdr->e_phoff)); \
+	printf("\t.e_shoff     = %-10"PRIi64" , /* (bytes into file) */\n", EGET(ehdr->e_shoff)); \
+	printf("\t.e_flags     = 0x%-8X ,\n", (uint32_t)EGET(ehdr->e_flags)); \
+	printf("\t.e_ehsize    = %-10u , /* (bytes) */\n", (uint16_t)EGET(ehdr->e_ehsize)); \
+	printf("\t.e_phentsize = %-10u , /* (bytes) */\n", (uint16_t)EGET(ehdr->e_phentsize)); \
 	/* TODO: Handle PN_XNUM */ \
-	printf("\t.e_phnum     = %-10i , /* (program headers) */\n", (int)EGET(ehdr->e_phnum)); \
-	printf("\t.e_shentsize = %-10i , /* (bytes) */\n", (int)EGET(ehdr->e_shentsize)); \
-	printf("\t.e_shnum     = %-10i , /* (section headers) */\n", (int)EGET(ehdr->e_shnum)); \
-	printf("\t.e_shstrndx  = %-10i\n", (int)EGET(ehdr->e_shstrndx)); \
+	printf("\t.e_phnum     = %-10u , /* (program headers) */\n", (uint16_t)EGET(ehdr->e_phnum)); \
+	printf("\t.e_shentsize = %-10u , /* (bytes) */\n", (uint16_t)EGET(ehdr->e_shentsize)); \
+	printf("\t.e_shnum     = %-10u , /* (section headers) */\n", (uint16_t)EGET(ehdr->e_shnum)); \
+	printf("\t.e_shstrndx  = %-10u\n", (uint16_t)EGET(ehdr->e_shstrndx)); \
 	printf("},\n"); \
 	}
 	DUMP_EHDR(32)
 	DUMP_EHDR(64)
 }
 
+static void dump_notes(elfobj *elf, size_t B, const void *memory, const void *memory_end)
+{
+	/* While normally we'd worry about Elf32_Nhdr vs Elf64_Nhdr, in the ELF
+	 * world, the two structs are exactly the same.  So avoid ugly CPP.
+	 */
+	size_t i;
+	const void *ndata = memory;
+	const char *name;
+	const unsigned char *desc;
+	uint32_t namesz, descsz;
+	const Elf32_Nhdr *note;
+	/* The first few bytes are the same between 32 & 64 bit ELFs. */
+	uint16_t e_type = EGET(((const Elf32_Ehdr *)elf->ehdr)->e_type);
+
+	if (memory_end > elf->data_end) {
+		printf("\n\t/%c note section is corrupt */\n", '*');
+		return;
+	}
+
+	printf("\n\t/%c note section dump:\n", '*');
+	for (i = 0; ndata < memory_end; ++i) {
+		note = ndata;
+		namesz = EGET(note->n_namesz);
+		descsz = EGET(note->n_descsz);
+		name = namesz ? ndata + sizeof(*note) : "";
+		desc = descsz ? ndata + sizeof(*note) + ALIGN_UP(namesz, 4) : "";
+		ndata += sizeof(*note) + ALIGN_UP(namesz, 4) + ALIGN_UP(descsz, 4);
+
+		if (ndata > memory_end) {
+			printf("\tNote is corrupt\n");
+			break;
+		}
+
+		printf("\t * Elf%zu_Nhdr note%zu = {\n", B, i);
+		printf("\t * \t.n_namesz = %u, (bytes) [%s]\n", namesz, name);
+		printf("\t * \t.n_descsz = %u, (bytes)", descsz);
+		if (descsz) {
+			printf(" [ ");
+			for (i = 0; i < descsz; ++i)
+				printf("%.2X ", desc[i]);
+			printf("]");
+		}
+		printf("\n");
+		printf("\t * \t.n_type   = %"PRIX64", [%s]\n",
+		       EGET(note->n_type), get_elfnttype(e_type, name, EGET(note->n_type)));
+		printf("\t * };\n");
+	}
+	printf("\t */\n");
+}
+
 static const char *dump_p_flags(uint32_t type, uint32_t flags)
 {
 	static char buf[1024];
@@ -235,39 +285,72 @@ static const char *dump_p_flags(uint32_t type, uint32_t flags)
 
 	return buf + 3;
 }
-static void dump_phdr(elfobj *elf, const void *phdr_void, long phdr_cnt)
+static void dump_phdr(elfobj *elf, const void *phdr_void, size_t phdr_cnt)
 {
 #define DUMP_PHDR(B) \
 	if (elf->elf_class == ELFCLASS ## B) { \
 	const Elf ## B ## _Phdr *phdr = PHDR ## B (phdr_void); \
+	Elf ## B ## _Off offset = EGET(phdr->p_offset); \
+	void *vdata = elf->vdata + offset; \
 	uint32_t p_type = EGET(phdr->p_type); \
 	switch (p_type) { \
 	case PT_DYNAMIC: phdr_dynamic_void = phdr_void; break; \
 	} \
-	printf("/* Program Header #%li 0x%lX */\n{\n", phdr_cnt, (uintptr_t)phdr_void - (uintptr_t)elf->data); \
-	printf("\t.p_type   = %-10li , /* [%s] */\n", (long)p_type, get_elfptype(p_type)); \
-	printf("\t.p_offset = %-10li , /* (bytes into file) */\n", (long)EGET(phdr->p_offset)); \
-	printf("\t.p_vaddr  = 0x%-8lX , /* (virtual addr at runtime) */\n", (unsigned long)EGET(phdr->p_vaddr)); \
-	printf("\t.p_paddr  = 0x%-8lX , /* (physical addr at runtime) */\n", (unsigned long)EGET(phdr->p_paddr)); \
-	printf("\t.p_filesz = %-10li , /* (bytes in file) */\n", (long)EGET(phdr->p_filesz)); \
-	printf("\t.p_memsz  = %-10li , /* (bytes in mem at runtime) */\n", (long)EGET(phdr->p_memsz)); \
-	printf("\t.p_flags  = 0x%-8lX , /* %s */\n", (unsigned long)EGET(phdr->p_flags), dump_p_flags(p_type, EGET(phdr->p_flags))); \
-	printf("\t.p_align  = %-10li , /* (min mem alignment in bytes) */\n", (long)EGET(phdr->p_align)); \
+	printf("/* Program Header #%zu 0x%tX */\n{\n", \
+	       phdr_cnt, (uintptr_t)phdr_void - elf->udata); \
+	printf("\t.p_type   = %-10u , /* [%s] */\n", p_type, get_elfptype(p_type)); \
+	printf("\t.p_offset = %-10"PRIi64" , /* (bytes into file) */\n", EGET(phdr->p_offset)); \
+	printf("\t.p_vaddr  = 0x%-8"PRIX64" , /* (virtual addr at runtime) */\n", EGET(phdr->p_vaddr)); \
+	printf("\t.p_paddr  = 0x%-8"PRIX64" , /* (physical addr at runtime) */\n", EGET(phdr->p_paddr)); \
+	printf("\t.p_filesz = %-10"PRIu64" , /* (bytes in file) */\n", EGET(phdr->p_filesz)); \
+	printf("\t.p_memsz  = %-10"PRIu64" , /* (bytes in mem at runtime) */\n", EGET(phdr->p_memsz)); \
+	printf("\t.p_flags  = 0x%-8X , /* %s */\n", (uint32_t)EGET(phdr->p_flags), dump_p_flags(p_type, EGET(phdr->p_flags))); \
+	printf("\t.p_align  = %-10"PRIu64" , /* (min mem alignment in bytes) */\n", EGET(phdr->p_align)); \
+	\
+	if ((off_t)EGET(phdr->p_offset) > elf->len) { \
+		printf("\t/* Warning: Program segment is corrupt. */\n"); \
+		goto done##B; \
+	} \
+	\
+	switch (p_type) { \
+	case PT_NOTE: \
+		dump_notes(elf, B, vdata, vdata + EGET(phdr->p_filesz)); \
+		break; \
+	} \
+ done##B: \
 	printf("},\n"); \
 	}
 	DUMP_PHDR(32)
 	DUMP_PHDR(64)
 }
 
-static void dump_shdr(elfobj *elf, const void *shdr_void, long shdr_cnt, const char *name)
+static const char *timestamp(uint64_t stamp)
 {
-	unsigned long i;
+	/* This doesn't work when run on a 32-bit host with 32-bit time_t beyond
+	 * beyond 2038, but we'll worry about that later.
+	 */
+	static char buf[20];
+	time_t t;
+	struct tm *tm;
+
+	t = stamp;
+	tm = gmtime(&t);
+	snprintf (buf, sizeof(buf), "%04u-%02u-%02u %02u:%02u:%02u",
+		tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
+		tm->tm_hour, tm->tm_min, tm->tm_sec);
+
+	return buf;
+}
+
+static void dump_shdr(elfobj *elf, const void *shdr_void, size_t shdr_cnt, const char *section_name)
+{
+	size_t i;
 
 	/* Make sure the string is valid. */
-	if ((void *)name >= elf->data_end)
-		name = "<corrupt>";
-	else if (memchr(name, 0, elf->len - (name - elf->data)) == NULL)
-		name = "<corrupt>";
+	if ((void *)section_name >= elf->data_end)
+		section_name = "<corrupt>";
+	else if (memchr(section_name, 0, elf->len - (section_name - elf->data)) == NULL)
+		section_name = "<corrupt>";
 
 #define DUMP_SHDR(B) \
 	if (elf->elf_class == ELFCLASS ## B) { \
@@ -276,18 +359,18 @@ static void dump_shdr(elfobj *elf, const void *shdr_void, long shdr_cnt, const c
 	uint32_t type = EGET(shdr->sh_type); \
 	uint ## B ## _t size = EGET(shdr->sh_size); \
 	\
-	printf("/* Section Header #%li '%s' 0x%lX */\n{\n", \
-	       shdr_cnt, name, (uintptr_t)shdr_void - (uintptr_t)elf->data); \
-	printf("\t.sh_name      = %-10i ,\n", (int)EGET(shdr->sh_name)); \
-	printf("\t.sh_type      = %-10i , /* [%s] */\n", (int)EGET(shdr->sh_type), get_elfshttype(type)); \
-	printf("\t.sh_flags     = %-10li ,\n", (long)EGET(shdr->sh_flags)); \
-	printf("\t.sh_addr      = 0x%-8lX ,\n", (unsigned long)EGET(shdr->sh_addr)); \
-	printf("\t.sh_offset    = %-10li , /* (bytes) */\n", (long)offset); \
-	printf("\t.sh_size      = %-10lu , /* (bytes) */\n", (unsigned long)size); \
-	printf("\t.sh_link      = %-10i ,\n", (int)EGET(shdr->sh_link)); \
-	printf("\t.sh_info      = %-10i ,\n", (int)EGET(shdr->sh_info)); \
-	printf("\t.sh_addralign = %-10li ,\n", (long)EGET(shdr->sh_addralign)); \
-	printf("\t.sh_entsize   = %-10li\n", (long)EGET(shdr->sh_entsize)); \
+	printf("/* Section Header #%zu '%s' 0x%tX */\n{\n", \
+	       shdr_cnt, section_name, (uintptr_t)shdr_void - elf->udata); \
+	printf("\t.sh_name      = %-10u ,\n", (uint32_t)EGET(shdr->sh_name)); \
+	printf("\t.sh_type      = %-10u , /* [%s] */\n", (uint32_t)EGET(shdr->sh_type), get_elfshttype(type)); \
+	printf("\t.sh_flags     = %-10"PRIu64" ,\n", EGET(shdr->sh_flags)); \
+	printf("\t.sh_addr      = 0x%-8"PRIX64" ,\n", EGET(shdr->sh_addr)); \
+	printf("\t.sh_offset    = %-10"PRIi64" , /* (bytes) */\n", (uint64_t)offset); \
+	printf("\t.sh_size      = %-10"PRIu64" , /* (bytes) */\n", (uint64_t)size); \
+	printf("\t.sh_link      = %-10u ,\n", (uint32_t)EGET(shdr->sh_link)); \
+	printf("\t.sh_info      = %-10u ,\n", (uint32_t)EGET(shdr->sh_info)); \
+	printf("\t.sh_addralign = %-10"PRIu64" ,\n", (uint64_t)EGET(shdr->sh_addralign)); \
+	printf("\t.sh_entsize   = %-10"PRIu64"\n", (uint64_t)EGET(shdr->sh_entsize)); \
 	\
 	if (type == SHT_NOBITS) { \
 		/* Special case so we can do valid check next. */ \
@@ -300,11 +383,11 @@ static void dump_shdr(elfobj *elf, const void *shdr_void, long shdr_cnt, const c
 		unsigned char *data = vdata; \
 		switch (type) { \
 		case SHT_PROGBITS: { \
-			if (strcmp(name, ".interp") == 0) { \
+			if (strcmp(section_name, ".interp") == 0) { \
 				printf("\n\t/* ELF interpreter: %s */\n", data); \
 				break; \
 			} \
-			if (strcmp(name, ".comment") != 0) \
+			if (strcmp(section_name, ".comment") != 0) \
 				break; \
 			break; \
 		} \
@@ -331,19 +414,38 @@ static void dump_shdr(elfobj *elf, const void *shdr_void, long shdr_cnt, const c
 			Elf##B##_Sym *sym = vdata; \
 			printf("\n\t/%c section dump:\n", '*'); \
 			for (i = 0; i < EGET(shdr->sh_size) / EGET(shdr->sh_entsize); ++i) { \
-				printf("\t * Elf%i_Sym sym%li = {\n", B, (long)i); \
-				printf("\t * \t.st_name  = %i,\n", (unsigned int)EGET(sym->st_name)); \
-				printf("\t * \t.st_value = 0x%lX,\n", (unsigned long)EGET(sym->st_value)); \
-				printf("\t * \t.st_size  = %li, (bytes)\n", (unsigned long)EGET(sym->st_size)); \
-				printf("\t * \t.st_info  = %i,\n", (unsigned int)EGET(sym->st_info)); \
-				printf("\t * \t.st_other = %i,\n", (unsigned int)EGET(sym->st_other)); \
-				printf("\t * \t.st_shndx = %li\n", (unsigned long)EGET(sym->st_shndx)); \
+				printf("\t * Elf%i_Sym sym%zu = {\n", B, i); \
+				printf("\t * \t.st_name  = %u,\n", (uint32_t)EGET(sym->st_name)); \
+				printf("\t * \t.st_value = 0x%"PRIX64",\n", EGET(sym->st_value)); \
+				printf("\t * \t.st_size  = %"PRIu64", (bytes)\n", EGET(sym->st_size)); \
+				printf("\t * \t.st_info  = %u,\n", (unsigned char)EGET(sym->st_info)); \
+				printf("\t * \t.st_other = %u,\n", (unsigned char)EGET(sym->st_other)); \
+				printf("\t * \t.st_shndx = %u\n", (uint16_t)EGET(sym->st_shndx)); \
 				printf("\t * };\n"); \
 				++sym; \
 			} \
 			printf("\t */\n"); \
 			break; \
 		} \
+		case SHT_NOTE: \
+			dump_notes(elf, B, vdata, vdata + EGET(shdr->sh_size)); \
+			break; \
+		case SHT_GNU_LIBLIST: { \
+			Elf##B##_Lib *lib = vdata; \
+			printf("\n\t/%c section dump:\n", '*'); \
+			for (i = 0; i < EGET(shdr->sh_size) / EGET(shdr->sh_entsize); ++i) { \
+				printf("\t * Elf%i_Lib lib%zu = {\n", B, i); \
+				printf("\t * \t.l_name       = %"PRIu64",\n", EGET(lib->l_name)); \
+				printf("\t * \t.l_time_stamp = 0x%"PRIX64", (%s)\n", \
+				       EGET(lib->l_time_stamp), timestamp(EGET(lib->l_time_stamp))); \
+				printf("\t * \t.l_checksum   = 0x%"PRIX64",\n", EGET(lib->l_checksum)); \
+				printf("\t * \t.l_version    = %"PRIu64",\n", EGET(lib->l_version)); \
+				printf("\t * \t.l_flags      = 0x%"PRIX64"\n", EGET(lib->l_flags)); \
+				printf("\t * };\n"); \
+				++lib; \
+			} \
+			printf("\t */\n"); \
+		} \
 		default: { \
 			if (be_verbose <= 1) \
 				break; \
@@ -365,18 +467,18 @@ static void dump_shdr(elfobj *elf, const void *shdr_void, long shdr_cnt, const c
 	DUMP_SHDR(64)
 }
 
-static void dump_dyn(elfobj *elf, const void *dyn_void, long dyn_cnt)
+static void dump_dyn(elfobj *elf, const void *dyn_void, size_t dyn_cnt)
 {
 #define DUMP_DYN(B) \
 	if (elf->elf_class == ELFCLASS ## B) { \
 	const Elf ## B ## _Dyn *dyn = dyn_void; \
-	unsigned long tag = EGET(dyn->d_tag); \
-	printf("/* Dynamic tag #%li '%s' 0x%lX */\n{\n", \
-	       dyn_cnt, get_elfdtype(tag), (uintptr_t)dyn_void - (uintptr_t)elf->data); \
-	printf("\t.d_tag     = 0x%-8lX ,\n", tag); \
+	int64_t tag = EGET(dyn->d_tag); \
+	printf("/* Dynamic tag #%zu '%s' 0x%tX */\n{\n", \
+	       dyn_cnt, get_elfdtype(tag), (uintptr_t)dyn_void - elf->udata); \
+	printf("\t.d_tag     = 0x%-8"PRIX64" ,\n", tag); \
 	printf("\t.d_un      = {\n"); \
-	printf("\t\t.d_val = 0x%-8lX ,\n", (unsigned long)EGET(dyn->d_un.d_val)); \
-	printf("\t\t.d_ptr = 0x%-8lX ,\n", (unsigned long)EGET(dyn->d_un.d_val)); \
+	printf("\t\t.d_val = 0x%-8"PRIX64" ,\n", EGET(dyn->d_un.d_val)); \
+	printf("\t\t.d_ptr = 0x%-8"PRIX64" ,\n", EGET(dyn->d_un.d_val)); \
 	printf("\t},\n"); \
 	printf("},\n"); \
 	}
@@ -403,7 +505,7 @@ static const char * const opts_help[] = {
 /* display usage and exit */
 static void usage(int status)
 {
-	int i;
+	size_t i;
 	printf("* Dump internal ELF structure\n\n"
 	       "Usage: %s <file1> [file2 fileN ...]\n\n", argv0);
 	printf("Options:\n");
@@ -449,7 +551,7 @@ static void parseargs(int argc, char *argv[])
 		err("Nothing to dump !?");
 
 	{
-	long file_cnt = 0;
+	size_t file_cnt = 0;
 
 	while (optind < argc)
 		dumpelf(argv[optind++], file_cnt++);
